//
// LDASTools Generic API - A library of general use algorithms for LDASTools
// Suite
//
// Copyright (C) 2023 California Institute of Technology
//
// LDASTools Generic API is free software; you may redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 (GPLv2) of the
// License or at your discretion, any later version.
//
// LDASTools Generic API is distributed in the hope that it will be useful, but
// without any warranty or even the implied warranty of merchantability
// or fitness for a particular purpose. See the GNU General Public
// License (GPLv2) for more details.
//
// Neither the names of the California Institute of Technology (Caltech),
// The Massachusetts Institute of Technology (M.I.T), The Laser
// Interferometer Gravitational-Wave Observatory (LIGO), nor the names
// of its contributors may be used to endorse or promote products derived
// from this software without specific prior written permission.
//
// You should have received a copy of the licensing terms for this
// software included in the file LICENSE located in the top-level
// directory of this package. If you did not, you can view a copy at
// http://dcc.ligo.org/M1500244/LICENSE
//

#include "genericAPI/Logging.hh"
#include "genericAPI/LogText.hh"

#include "frameAPI/createRDS.hh"

#define BOOST_TEST_MAIN
#include <boost/test/included/unit_test.hpp>
#include <boost/algorithm/string/join.hpp>

struct rds_global_fixture
{
    rds_global_fixture( )
    {
        //-------------------------------------------------------------------
        // Setup logging as is needed by query class
        //-------------------------------------------------------------------
        GenericAPI::SetLogFormatter( new GenericAPI::Log::Text( "" ) );
        static char cwd_buffer[ 2048 ];
        if ( getcwd( cwd_buffer,
                     sizeof( cwd_buffer ) / sizeof( *cwd_buffer ) ) ==
             (const char*)NULL )
        {
            exit( 1 );
        }
        std::string                       cwd( cwd_buffer );
        GenericAPI::Log::stream_file_type fs( new GenericAPI::Log::StreamFile );
        LDASTools::AL::Log::stream_type   s;

        s = fs;
        fs->FilenameExtension( GenericAPI::LogFormatter( )->FileExtension( ) );
        GenericAPI::LogFormatter( )->Stream( s );
        GenericAPI::setLogTag( "test_query" );
        GenericAPI::LoggingInfo::LogDirectory( cwd );

        std::cout << "global setup" << std::endl;
    }

    ~rds_global_fixture( )
    {
        std::cout << "global teardown" << std::endl;
    }
};

BOOST_GLOBAL_FIXTURE( rds_global_fixture );

BOOST_AUTO_TEST_CASE( channel_simple )
{
    static const int         size_expected = 1;
    static const std::string test_channel_string( "H1:channel1" // channel name
    );
    static const std::string name_expected( "H1:channel1" );
    static const int         resample_expected = 1;
    static const std::string rename_expected( "H1:channel1" );

    FrameAPI::channel_container_type channels;
    channels.ParseChannelName( test_channel_string );

    BOOST_CHECK_MESSAGE( channels.names.size( ) == size_expected,
                         "channels.names.size( ) is '"
                             << channels.names.size( ) << "' instead of '"
                             << size_expected << "'" );
    BOOST_CHECK_MESSAGE( channels.resampling.size( ) == size_expected,
                         "channels.resampling.size( ) is '"
                             << channels.resampling.size( ) << "' instead of '"
                             << size_expected << "'" );
    if ( ( channels.names.size( ) == 1 ) &&
         ( channels.resampling.size( ) == 1 ) )
    {
        BOOST_CHECK_MESSAGE( channels.names.front( ).old_channel_name == name_expected,
                             "channels.names.front( ).old_channel_name is '"
                                 << channels.names.front( ).old_channel_name << "' instead of '"
                                 << name_expected << "'" );
        BOOST_CHECK_MESSAGE( channels.names.front( ).new_channel_name == rename_expected,
                             "channels.names.front( ).new_channel_name is '"
                             << channels.names.front( ).new_channel_name
                             << "' instead of '" << rename_expected
                             << "'" );
        BOOST_CHECK_MESSAGE( channels.resampling.front( ) == resample_expected,
                             "channels.resampling.front( ) is '"
                                 << channels.resampling.front( )
                                 << "' instead of '" << resample_expected
                                 << "'" );
    }
}

BOOST_AUTO_TEST_CASE( channel_resample )
{
    static const int         size_expected = 1;
    static const std::string test_channel_string( "H1:channel1!4" // channel name
    );
    static const std::string name_expected( "H1:channel1" );
    static const int         resample_expected = 4;
    static const std::string rename_expected( "H1:channel1" );

    FrameAPI::channel_container_type channels;
    channels.ParseChannelName( test_channel_string );

    BOOST_CHECK_MESSAGE( channels.names.size( ) == size_expected,
                         "channels.names.size( ) is '"
                             << channels.names.size( ) << "' instead of '"
                             << size_expected << "'" );
    BOOST_CHECK_MESSAGE( channels.resampling.size( ) == size_expected,
                         "channels.resampling.size( ) is '"
                             << channels.resampling.size( ) << "' instead of '"
                             << size_expected << "'" );
    if ( ( channels.names.size( ) == 1 ) &&
         ( channels.resampling.size( ) == 1 ) )
    {
        BOOST_CHECK_MESSAGE( channels.names.front( ).old_channel_name == name_expected,
                             "channels.names.front( ).old_channel_name is '"
                                 << channels.names.front( ).old_channel_name << "' instead of '"
                                 << name_expected << "'" );
        BOOST_CHECK_MESSAGE( channels.names.front( ).new_channel_name == rename_expected,
                             "channels.names.front( ).new_channel_name is '"
                             << channels.names.front( ).new_channel_name
                             << "' instead of '" << rename_expected
                             << "'" );
        BOOST_CHECK_MESSAGE( channels.resampling.front( ) == resample_expected,
                             "channels.resampling.front( ) is '"
                                 << channels.resampling.front( )
                                 << "' instead of '" << resample_expected
                                 << "'" );
    }
}

BOOST_AUTO_TEST_CASE( channel_rename )
{
    static const int         size_expected = 1;
    static const std::string test_channel_string( "H1:channel1=H1:channel1_AR" // channel name
    );
    static const std::string name_expected( "H1:channel1" );
    static const int         resample_expected = 1;
    static const std::string rename_expected( "H1:channel1_AR" );

    FrameAPI::channel_container_type channels;
    channels.ParseChannelName( test_channel_string );

    BOOST_CHECK_MESSAGE( channels.names.size( ) == size_expected,
                         "channels.names.size( ) is '"
                             << channels.names.size( ) << "' instead of '"
                             << size_expected << "'" );
    BOOST_CHECK_MESSAGE( channels.resampling.size( ) == size_expected,
                         "channels.resampling.size( ) is '"
                             << channels.resampling.size( ) << "' instead of '"
                             << size_expected << "'" );
    if ( ( channels.names.size( ) == 1 ) &&
         ( channels.resampling.size( ) == 1 ) )
    {
        BOOST_CHECK_MESSAGE( channels.names.front( ).old_channel_name == name_expected,
                             "channels.names.front( ).old_channel_name is '"
                                 << channels.names.front( ).old_channel_name << "' instead of '"
                                 << name_expected << "'" );
        BOOST_CHECK_MESSAGE( channels.names.front( ).new_channel_name == rename_expected,
                             "channels.names.front( ).new_channel_name is '"
                             << channels.names.front( ).new_channel_name
                             << "' instead of '" << rename_expected
                             << "'" );
        BOOST_CHECK_MESSAGE( channels.resampling.front( ) == resample_expected,
                             "channels.resampling.front( ) is '"
                                 << channels.resampling.front( )
                                 << "' instead of '" << resample_expected
                                 << "'" );
    }
}

BOOST_AUTO_TEST_CASE( channel_resample_and_rename )
{
    static const int         size_expected = 1;
    static const std::string test_channel_string( "H1:channel1!4=H1:channel1_AR" // channel name
    );
    static const std::string name_expected( "H1:channel1" );
    static const int         resample_expected = 4;
    static const std::string rename_expected( "H1:channel1_AR" );

    FrameAPI::channel_container_type channels;
    channels.ParseChannelName( test_channel_string );

    BOOST_CHECK_MESSAGE( channels.names.size( ) == size_expected,
                         "channels.names.size( ) is '"
                             << channels.names.size( ) << "' instead of '"
                             << size_expected << "'" );
    BOOST_CHECK_MESSAGE( channels.resampling.size( ) == size_expected,
                         "channels.resampling.size( ) is '"
                             << channels.resampling.size( ) << "' instead of '"
                             << size_expected << "'" );
    if ( ( channels.names.size( ) == 1 ) &&
         ( channels.resampling.size( ) == 1 ) )
    {
        BOOST_CHECK_MESSAGE( channels.names.front( ).old_channel_name == name_expected,
                             "channels.names.front( ).old_channel_name is '"
                                 << channels.names.front( ).old_channel_name << "' instead of '"
                                 << name_expected << "'" );
        BOOST_CHECK_MESSAGE( channels.names.front( ).new_channel_name == rename_expected,
                             "channels.names.front( ).new_channel_name is '"
                                 << channels.names.front( ).new_channel_name
                                 << "' instead of '" << rename_expected
                                 << "'" );
        BOOST_CHECK_MESSAGE( channels.resampling.front( ) == resample_expected,
                             "channels.resampling.front( ) is '"
                             << channels.resampling.front( )
                             << "' instead of '" << resample_expected
                             << "'" );
    }
}
